"""
Magics to have disqus conversation in the notebook.
"""

from __future__ import annotations

from IPython.display import HTML

from metakernel import Magic, MetaKernel


class ConversationMagic(Magic):
    def cell_conversation(self, id: str) -> None:
        """
        %conversation ID - insert conversation by ID
        %%conversation ID - insert conversation by ID
        """
        html = f"""
<div id="disqus_thread"></div>
<script>
(function() {{ // DON'T EDIT BELOW THIS LINE
    var d = document, s = d.createElement('script');
    s.src = '//{id}.disqus.com/embed.js';
    s.setAttribute('data-timestamp', +new Date());
    (d.head || d.body).appendChild(s);
}})();
</script>
<noscript>Please enable JavaScript to view the <a href="https://disqus.com/?ref_noscript">comments powered by Disqus.</a></noscript>
"""
        self.kernel.Display(HTML(html))  # type: ignore[no-untyped-call]

    def line_conversation(self, id: str) -> None:
        """
        %conversation ID - insert conversation by ID
        %%conversation ID - insert conversation by ID
        """
        self.cell_conversation(id)
        self.evaluate = False


def register_magics(kernel: MetaKernel) -> None:
    kernel.register_magics(ConversationMagic)


def register_ipython_magics() -> None:
    from metakernel import IPythonKernel
    from metakernel.magic import register_cell_magic, register_line_magic

    kernel = IPythonKernel()
    magic = ConversationMagic(kernel)

    @register_line_magic
    def conversation(id: str) -> None:
        magic.line_conversation(id)

    @register_cell_magic  # type: ignore[no-redef]
    def conversation(id: str, cell: str) -> None:  # noqa: F811
        magic.code = cell
        magic.cell_conversation(id)



from typing import Any

from IPython.display import clear_output as ipclear_output
from IPython.display import display as ipdisplay


def display(*args: Any, **kwargs: Any) -> None:
    from . import get_metakernel

    kernel = get_metakernel()
    if kernel:
        kernel.Display(*args, **kwargs)
    else:
        ipdisplay(*args, **kwargs)  # type:ignore[no-untyped-call]


def clear_output(*args: Any, **kwargs: Any) -> None:
    from . import get_metakernel

    kernel = get_metakernel()
    if kernel:
        wait = kwargs.get("wait", False)
        kernel.send_response(kernel.iopub_socket, "clear_output", {"wait": wait})
    else:
        ipclear_output(*args, **kwargs)  # type:ignore[no-untyped-call]



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import ast
import pydoc
import sys
from typing import Any

import jedi  # type:ignore[import-untyped]
from jedi import Interpreter
from jedi.api.helpers import get_on_completion_name  # type:ignore[import-untyped]
from parso import split_lines  # type:ignore[attr-defined]

from metakernel import ExceptionWrapper, Magic, MetaKernel, option


def exec_then_eval(code: str, env: dict[str, Any]) -> Any:
    import traceback

    try:
        block = ast.parse(code, mode="exec")
        last = block.body.pop()
        if not isinstance(last, ast.Expr):
            block.body.append(last)
            exec(compile(block, "python cell", mode="exec"), env)
            retval = None
        else:
            exec(compile(block, "python cell", mode="exec"), env)
            retval = eval(
                compile(ast.Expression(last.value), "python cell", mode="eval"), env
            )
        if "retval" in env and env["retval"] is not None:
            return env["retval"]
        else:
            return retval

    except Exception as exc:
        ex_type, ex, tb = sys.exc_info()
        line1 = ["Traceback (most recent call last):"]
        ex_name = ex_type.__name__ if ex_type is not None else "Exception"
        line2 = [f"{ex_name}: {ex!s}"]
        tb_format = (
            line1 + [line.rstrip() for line in traceback.format_tb(tb)[1:]] + line2
        )
        return ExceptionWrapper(ex_name, repr(exc.args), tb_format)


class PythonMagic(Magic):
    def __init__(self, kernel: MetaKernel) -> None:
        super().__init__(kernel)
        self.env = globals()["__builtins__"].copy()
        self.retval: Any = None

    def line_python(self, *args: Any) -> None:
        """
        %python CODE - evaluate code as Python

        This line magic will evaluate the CODE (either expression or
        statement) as Python code.

        Note that the version of Python is that of the notebook server.

        Examples:
            %python x = 42
            %python import math
            %python x + math.pi

        """
        code = " ".join(args)
        self.retval = self.eval(code)
        self.env["retval"] = None

    def eval(self, code: str) -> Any:
        import IPython.display

        import metakernel
        import metakernel.display

        # Ensure the current kernel is the active one so that display() calls
        # inside %%python route to this kernel's output, not a stale instance.
        metakernel.MetaKernel.meta_kernel = self.kernel

        # monkey patch IPython.display.display
        # to redirect notebook display calls to kernel display
        IPython.display.display = metakernel.display.display

        if "__builtins__" not in self.env:
            ## __builtins__ get generated after an eval:
            eval("1", self.env)
            ## make 'kernel' and 'input' available:
            self.env["__builtins__"]["kernel"] = self.kernel
            self.env["__builtins__"]["input"] = self.kernel.raw_input
            self.env["input"] = self.kernel.raw_input
        return exec_then_eval(code.strip(), self.env)

    @option(
        "-e",
        "--eval_output",
        action="store_true",
        default=False,
        help="Use the retval value from the Python cell as code in the kernel language.",
    )
    def cell_python(self, eval_output: bool = False) -> None:
        """
        %%python - evaluate contents of cell as Python

        This cell magic will evaluate the cell (either expression or
        statement) as Python code.

        Unlike IPython's Python, this does not return the last expression.
        To do that, you need to assign the last expression to the special
        variable "retval".

        The -e or --eval_output flag signals that the retval value expression
        will be used as code for the cell to be evaluated by the host
        language.

        Note that the version of Python is that of the notebook server.

        Examples:
            %%python
            x = 42

            %%python
            import math
            retval = x + math.pi

            %%python -e
            retval = "'(this is code in the kernel language)"

            %%python -e
            "'(this is code in the kernel language)"

        """
        if self.code.strip():
            if eval_output:
                retval = self.eval(self.code)
                self.code = (
                    str(self.env["retval"])
                    if ("retval" in self.env and self.env["retval"] is not None)
                    else str(retval)
                )
                self.retval = None
                self.env["retval"] = None
                self.evaluate = True
            else:
                self.retval = self.eval(self.code)
                self.env["retval"] = None
                self.evaluate = False

    def post_process(self, retval: Any) -> Any:
        if retval is not None:
            return retval
        else:
            return self.retval

    def get_completions(self, info: dict[str, Any]) -> list[str]:
        """Get Python completions"""
        # https://github.com/davidhalter/jedi/blob/master/jedi/utils.py
        if jedi is None:
            return []

        text = info["code"]
        position = (info["line_num"], info["column"])
        interpreter = Interpreter(text, [self.env])

        lines = split_lines(text)
        name = get_on_completion_name(interpreter._module_node, lines, position)
        before = text[: len(text) - len(name)]

        try:
            completions = interpreter.complete()
        except AttributeError:
            completions = interpreter.completions()

        completions = [before + c.name_with_symbols for c in completions]

        self.kernel.log.error(completions)

        return [c[info["start"] :] for c in completions]

    def get_help_on(
        self, info: dict[str, Any], level: int = 0, none_on_fail: bool = False
    ) -> str | None:
        """Implement basic help for functions"""
        if not info["code"]:
            return None if none_on_fail else ""

        last = info["obj"]

        default = None if none_on_fail else (f'No help available for "{last}"')

        parts = last.split(".")

        obj = self.env.get(parts[0], None)

        if not obj:
            return default

        for p in parts[1:]:
            obj = getattr(obj, p, None)

            if not obj:
                return default

        strhelp = pydoc.render_doc(obj, "Help on %s")
        if level == 0:
            return getattr(obj, "__doc__", strhelp)
        else:
            return strhelp


def register_magics(kernel: MetaKernel) -> None:
    kernel.register_magics(PythonMagic)



from __future__ import annotations

import atexit
import errno
import os
import signal
import sys
import time
from typing import Any

from . import pexpect

PEXPECT_PROMPT = "PEXPECT_PROMPT>"
PEXPECT_STDIN_PROMPT = "PEXPECT_PROMPT+"
PEXPECT_CONTINUATION_PROMPT = "PEXPECT_PROMPT_"


class REPLWrapper:
    """Wrapper for a REPL.

    All prompts are interpreted as regexes.  If you have special
    characters in the prompt, use `re.escape` to escape the characters.

    :param cmd_or_spawn: This can either be an instance of
    :class:`pexpect.spawn` in which a REPL has already been started,
    or a str command to start a new REPL process.
    :param str prompt_regex:  Regular expression representing process prompt, eg ">>>" in Python.
    :param str continuation_prompt_regex: Regular expression representing process continuation prompt, e.g. "..." in Python.
    :param str prompt_change_cmd: Optional kernel command that sets continuation-of-line-prompts, eg PS1 and PS2, such as "..." in Python.
        to something more unique. If this is ``None``, the prompt will not be
        changed. This will be formatted with the new and continuation prompts
        as positional parameters, so you can use ``{}`` style formatting to
        insert them into the command.
    :param str new_prompt_regex: The more unique prompt to expect after the change.
    :param str stdin_prompt_regex: The regex for a stdin prompt from the
        child process.  The prompt itself will be sent to the `stdin_handler`,
        so any sentinel value inserted will have to be removed by the caller.
    :param str extra_init_cmd: Commands to do extra initialisation, such as
      disabling pagers.
    :param str prompt_emit_cmd: Optional kernel command that emits the prompt
      when one is not emitted by default (typically happens on Windows only)
    :param bool force_prompt_on_continuation: Whether to force a prompt when
    we need to interrupt a continuation prompt.
    :param bool echo: Whether the child should echo, or in the case
    of Windows, whether the child does echo.
    :param dict extr_env: Extra env variables to send to the child.
    :param list args: Additional arguments to pass to the spawned command.
    """

    def __init__(
        self,
        cmd_or_spawn: Any,
        prompt_regex: Any,
        prompt_change_cmd: str | None,
        new_prompt_regex: Any = PEXPECT_PROMPT,
        continuation_prompt_regex: Any = PEXPECT_CONTINUATION_PROMPT,
        stdin_prompt_regex: Any = PEXPECT_STDIN_PROMPT,
        extra_init_cmd: str | None = None,
        prompt_emit_cmd: str | None = None,
        force_prompt_on_continuation: bool = False,
        echo: bool = False,
        extra_env: dict[str, Any] | None = None,
        encoding: str = "utf-8",
        args: list[str] | None = None,
    ) -> None:
        if isinstance(cmd_or_spawn, str):
            if extra_env:
                env = os.environ.copy()
                env.update(extra_env or {})
            else:
                env = None
            self.child = pexpect.spawnu(
                cmd_or_spawn,
                args=args,
                echo=echo,
                codec_errors="ignore",
                encoding=encoding,
                env=env,
            )
        else:
            self.child = cmd_or_spawn

        if self.child.echo and not echo:
            # Existing spawn instance has echo enabled, disable it
            # to prevent our input from being repeated to output.
            self.child.setecho(False)
            self.child.waitnoecho()

        self.echo = echo
        self.prompt_emit_cmd = prompt_emit_cmd
        self._force_prompt_on_continuation = force_prompt_on_continuation
        self.prompt_change_cmd: str | None = None
        self.prompt_regex: str | None = None

        if prompt_change_cmd is None:
            self.prompt_regex = prompt_regex
        else:
            formatted_prompt_change_cmd = prompt_change_cmd.format(
                new_prompt_regex, continuation_prompt_regex
            )
            self.set_prompt(prompt_regex, formatted_prompt_change_cmd)
            self.prompt_regex = new_prompt_regex
        self.continuation_prompt_regex = continuation_prompt_regex
        self.stdin_prompt_regex = stdin_prompt_regex

        self._stream_handler: Any = None
        self._stdin_handler: Any = None
        self._line_handler: Any = None

        self._expect_prompt()

        if extra_init_cmd is not None:
            self.run_command(extra_init_cmd)

        atexit.register(self.terminate)

    def sendline(self, line: str) -> None:
        self.child.sendline(line)
        if self.echo:
            self.child.readline()

    def set_prompt(self, prompt_regex: str, prompt_change_cmd: str) -> None:
        self.child.expect([r"[\s\S]+"])
        self.sendline(prompt_change_cmd)
        self.prompt_change_cmd = prompt_change_cmd

    def _expect_prompt(self, timeout: int | None = None) -> Any:
        """Expect a prompt from the child."""
        expects = [
            self.prompt_regex,
            self.continuation_prompt_regex,
            self.stdin_prompt_regex,
        ]
        if self.prompt_emit_cmd:
            self.sendline(self.prompt_emit_cmd)

        if self._stream_handler:
            return self._expect_prompt_stream(expects, timeout)

        if self._line_handler:
            expects += [self.child.crlf]

        while True:
            pos = self.child.expect(expects, timeout=timeout)
            # got a full prompt or continuation prompt.
            if pos in [0, 1]:
                return pos
            # got a stdin prompt
            if pos == 2:
                if not self._stdin_handler:
                    raise ValueError("Stdin Requested but not stdin handler available")

                resp = self._stdin_handler(self.child.after)
                self.sendline(resp)
            # got a newline
            else:
                self._line_handler(self.child.before.rstrip())

    def _expect_prompt_stream(
        self, expects: str | list[str | None], timeout: Any = None
    ) -> Any:
        """Expect a prompt with streaming output."""
        stream_handler = self._stream_handler
        stdin_handler = self._stdin_handler

        t0 = time.time()
        if timeout == -1:
            timeout = 30
        elif timeout is None:
            timeout = 1e6
        stream_timeout = 0.1

        unhandled_cr = False

        # Wait for a prompt, handling carriage returns.
        while True:
            if time.time() - t0 > timeout:
                raise pexpect.TIMEOUT("Timed out")

            try:
                # Wait for a prompt.
                pos = self.child.expect(expects, timeout=stream_timeout)
            except pexpect.TIMEOUT:
                # Process any carriage returns in the stream.
                while 1:
                    try:
                        self.child.expect(["\r"], timeout=0)
                        if unhandled_cr:
                            stream_handler("\r")
                        stream_handler(self.child.before)
                        unhandled_cr = True
                    except pexpect.TIMEOUT:
                        break
                continue

            # prompt or stdin request received, handle line.
            line = self.child.before

            # Handle cr state.
            if unhandled_cr:
                line = "\r" + line
            unhandled_cr = False

            # Handle stdin request.
            if pos == 2:
                if not stdin_handler:
                    raise ValueError("Stdin Requested but no stdin handler available")
                resp = stdin_handler(line + self.child.after)
                self.sendline(resp)
                continue

            # prompt received, but partial line precedes it
            if len(line) != 0:
                stream_handler(line)

            # exit on prompt received.
            break
        return pos

    def run_command(
        self,
        command: str,
        timeout: Any = None,
        stream_handler: Any = None,
        line_handler: Any = None,
        stdin_handler: Any = None,
    ) -> str:
        """Send a command to the REPL, wait for and return output.
        :param str command: The command to send. Trailing newlines are not needed.
          This should be a complete block of input that will trigger execution;
          if a continuation prompt is found after sending input, :exc:`ValueError`
          will be raised.
        :param int timeout: How long to wait for the next prompt. -1 means the
          default from the :class:`pexpect.spawn` object (default 30 seconds).
          None means to wait indefinitely.
        :param func stream_handler - A function that accepts a string to print as a streaming output
        :param func line_handler - A function that accepts a string to print as a line output
        :param stdin_handler - A function that prompts the user for input and
        returns a response.
        """
        # Split up multiline commands and feed them in bit-by-bit
        cmdlines = command.splitlines()
        # splitlines ignores trailing newlines - add it back in manually
        if command.endswith("\n"):
            cmdlines.append("")
        if not cmdlines:
            raise ValueError("No command was given")

        res = []
        self._line_handler = line_handler
        self._stream_handler = stream_handler
        self._stdin_handler = stdin_handler

        self.sendline(cmdlines[0])
        for line in cmdlines[1:]:
            if not self.prompt_emit_cmd:
                self._expect_prompt(timeout=timeout)
                res.append(strip_bracketing(self.child.before))
            self.sendline(line)

        # Command was fully submitted, now wait for the next prompt
        if self._expect_prompt(timeout=timeout) == 1:
            # Got a continuation prompt - could be a soft continuation (complete block
            # awaiting execution, e.g. `if True:\n    print('hi')`) or a hard
            # continuation (genuinely incomplete input, e.g. `if True:`).
            # Send an empty line to resolve soft continuations and check again.
            self.sendline("")
            if self._expect_prompt(timeout=timeout) == 1:
                # Still a continuation prompt - input was genuinely incomplete
                self.interrupt(continuation=True)
                raise ValueError(
                    "Continuation prompt found - input was incomplete:\n" + command
                )

        if self._stream_handler or self._line_handler:
            return ""
        return "".join(res + [strip_bracketing(self.child.before)])

    def interrupt(self, continuation: bool = False) -> Any:
        """Interrupt the process and wait for a prompt.

        Returns
        -------
        The value up to the prompt.
        """
        if pexpect.pty is not None:
            self.child.sendintr()
        else:
            self.child.kill(signal.SIGINT)  # type:ignore[unreachable]
        if continuation and self._force_prompt_on_continuation:
            self.sendline(self.prompt_change_cmd or "")
        while 1:
            try:
                self._expect_prompt(timeout=-1)
                break
            except KeyboardInterrupt:
                pass

        return self.child.before

    def terminate(self) -> Any:
        if pexpect.pty is not None:
            self.child.close()
            return self.child.terminate()
        try:  # type:ignore[unreachable]
            self.child.kill(signal.SIGTERM)
        except Exception as e:
            if e.errno != errno.EACCES:
                raise


def python(command: str = "python") -> REPLWrapper:
    """Start a Python shell and return a :class:`REPLWrapper` object."""
    if pexpect.pty is None:
        raise OSError(f'Not supported on platform "{sys.platform}"')
    return REPLWrapper(
        command,
        ">>> ",
        "import sys; sys.ps1={0!r}; sys.ps2={1!r}",
        extra_env=dict(PYTHON_BASIC_REPL="1"),
    )


def bash(command: str = "bash", prompt_regex: str = "[$#]") -> REPLWrapper:
    """Start a bash shell and return a :class:`REPLWrapper` object."""

    # If the user runs 'env', the value of PS1 will be in the output. To avoid
    # replwrap seeing that as the next prompt, we'll embed the marker characters
    # for invisible characters in the prompt; these show up when inspecting the
    # environment variable, but not when bash displays the prompt.
    ps1 = PEXPECT_PROMPT[:5] + r"\[\]" + PEXPECT_PROMPT[5:]
    ps2 = PEXPECT_CONTINUATION_PROMPT[:5] + r"\[\]" + PEXPECT_CONTINUATION_PROMPT[5:]
    prompt_change_cmd: str | None = f"PS0='' PS1='{ps1}' PS2='{ps2}' PROMPT_COMMAND=''"

    if os.name == "nt":
        prompt_regex = "__repl_ready__"
        prompt_emit_cmd = "echo __repl_ready__"
        prompt_change_cmd = None
    else:
        prompt_emit_cmd = None

    extra_init_cmd = "export TERM=dumb PAGER=cat"

    # Make sure the bash shell has a valid ending character.
    bashrc = os.path.join(os.path.dirname(pexpect.PEXPECT_DIR), "bashrc.sh")
    child = pexpect.spawn(command, ["--rcfile", bashrc], echo=False, encoding="utf-8")

    return REPLWrapper(
        child,
        prompt_regex,
        prompt_change_cmd,
        prompt_emit_cmd=prompt_emit_cmd,
        extra_init_cmd=extra_init_cmd,
    )


def powershell(command: str = "powershell", prompt_regex: str = ">") -> REPLWrapper:
    """ "Start a powershell and return a :class:`REPLWrapper` object."""
    return REPLWrapper(command, prompt_regex, 'Function prompt {{ "{0}" }}', echo=True)


def strip_bracketing(
    string: str, start: str = "\x1b[?2004l", stop: str = "\x1b[?2004h"
) -> str:
    """Strip 'bracketed paste' control characters, if they are present"""
    if string.startswith(start) and string.endswith(stop):
        return string[len(start) : -len(stop)]
    else:
        return string



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from typing import Any

from metakernel import Magic, MetaKernel, option

try:
    from calysto_scheme import scheme  # type: ignore[import-untyped]
except ImportError:
    scheme = None


class SchemeMagic(Magic):
    def __init__(self, kernel: MetaKernel) -> None:
        super().__init__(kernel)
        self.retval = None

    def line_scheme(self, *args: str) -> None:
        """
        %scheme CODE - evaluate code as Scheme

        This line magic will evaluate the CODE (either expression or
        statement) as Scheme code.

        Examples:
            %scheme (define x 42)
            %scheme (import "math")
            %scheme (+ x + math.pi)

        """
        code = " ".join(args)
        self.retval = self.eval(code)

    def eval(self, code: str) -> Any:
        if scheme:
            return scheme.execute_string_rm(code.strip())
        else:
            raise Exception("calysto_scheme is required")

    @option(
        "-e",
        "--eval_output",
        action="store_true",
        default=False,
        help="Use the retval value from the Scheme cell as code in the kernel language.",
    )
    def cell_scheme(self, eval_output: bool = False) -> None:
        """
        %%scheme - evaluate contents of cell as Scheme

        This cell magic will evaluate the cell (either expression or
        statement) as Scheme code.

        The -e or --eval_output flag signals that the retval value expression
        will be used as code for the cell to be evaluated by the host
        language.

        Examples:
            %%scheme
            (define x 42)

            %%scheme
            (import "math")
            (define retval (+ x math.pi))

            %%scheme -e
            (define retval "this = code")

            %%scheme -e
            "this = code"
        """
        if self.code.strip():
            if eval_output:
                self.code = self.eval(self.code)
                self.evaluate = True
            else:
                self.retval = self.eval(self.code)
                self.evaluate = False

    def post_process(self, retval: Any) -> Any:
        if retval is not None:
            return retval
        else:
            return self.retval


def register_magics(kernel: MetaKernel) -> None:
    kernel.register_magics(SchemeMagic)


def register_ipython_magics() -> None:
    from metakernel import IPythonKernel
    from metakernel.magic import register_cell_magic, register_line_magic

    kernel = IPythonKernel()
    magic = SchemeMagic(kernel)

    @register_line_magic
    def scheme(line: str) -> Any:
        magic.line_scheme(line)
        return magic.retval

    @register_cell_magic  # type: ignore[no-redef]
    def scheme(line: str, cell: str) -> Any:  # noqa: F811
        magic.code = cell
        magic.cell_scheme()
        return magic.retval
